import websocket
from synthetic import synthesize_property
import six
import json
import uuid
import time
import utils


class XoError(Exception):
    pass


class XoTimeoutError(Exception):
    pass


class XoApiError(XoError):
    def __init__(self, method, msg):
        self.method = method
        self.msg = msg


@synthesize_property('timeout', contract=int, default=10)
class xo:
    """ xo-server helper class """

    def __init__(self, server, email=None, password=None, token=None, learn_methods=True):
        if server is None:
            raise Exception()
        self._server = server + '/api/'
        self._email = email
        self._password = password
        self._token = token

        self._ws = websocket.WebSocket()
        self._ws.connect(self._server)

        if self._email is not None and self._password is not None:
            self.call('session.signInWithPassword', email=self._email, password=self._password)
        elif self._token is not None:
            self.call('session.signInWithToken', token=self._token)

        if learn_methods:
            self._learn_methods()

    def call(self, method, **kwargs):
        _uuid = uuid.uuid1()
        payload = {
            "method": method,
            "params": kwargs,
            "jsonrpc": "2.0",
            "id": _uuid.hex
        }

        self._ws.send(json.dumps(payload))
        start_time = time.time()
        resp = None
        while time.time() < (start_time + self.timeout):
            resp = json.loads(self._ws.recv())
            if ('id' not in resp) or resp['id'] != _uuid.hex:
                resp = None
                continue
            else:
                if 'error' in resp:
                    raise XoApiError(method, resp['error']['message'])
                break

        if resp is None:
            raise XoTimeoutError

        return resp['result']

    def _learn_methods(self):
        xoa_methods = self.call('system.getMethodsInfo')
        methods = list()
        for xoa_method_name, xoa_method_info in six.iteritems(xoa_methods):
            method_code = utils.render_function(xoa_method_name, xoa_method_info['params'])
            methods.append(method_code)
        for method_code in methods:
            utils.patch_method_in(self, method_code)
        return methods



import types
import six


def patch_method_in(obj, method_code):
    ld = dict()
    exec(method_code, None, ld)
    for name in ld:
        setattr(obj, name, types.MethodType(ld[name], obj))


def render_function(xoa_method_name, xoa_method_params):
    function_name = xoa_method_name.replace('.', '_')
    if six.PY2:
        func_template = """def {function_name}(self, **kwargs): return self.call('{xoa_method}', **kwargs)"""
        return func_template.format(function_name=function_name, xoa_method=xoa_method_name)
    else:
        func_template = """def {function_name}(self{sig_args}): return self.call('{xoa_name}'{proxy_args})"""
        sig_args = str()
        proxy_args = str()
        if len(xoa_method_params) > 0:
            sig_args += ', *'
            for param_name, param_attrs in xoa_method_params.items():
                if 'optional' in param_attrs and param_attrs['optional']:
                    sig_args += ', ' + str(param_name) + '=None'
                else:
                    sig_args += ', ' + str(param_name)
                proxy_args += ', ' + str(param_name) + '=' + str(param_name)
        return func_template.format(function_name=function_name, xoa_name=xoa_method_name,
                                    sig_args=sig_args, proxy_args=proxy_args)



from setuptools import setup

setup(
    name='xolib',
    version='0.2.1',
    author='Vyronas Tsingaras',
    author_email='vyronas@vtsingaras.me',
    packages=['xolib'],
    url='https://github.com/vtsingaras/python-xolib',
    download_url='https://github.com/vtsingaras/python-xolib/archive/v0.2.1.tar.gz',
    license='LICENSE.txt',
    description='Helper library for interfacing with xo-server.',
    long_description=open('README.rst').read(),
    install_requires=[
        "websocket-client==0.37.0",
        "six",
        "pysynthetic"
    ],
)
